from frozenlist2 import frozenlist

from pubtools.pulplib._impl import compat_attr as attr

from ..schema import load_schema
from .unit import Unit
from .common import PulpObject
from .attr import pulp_attrib


@attr.s(kw_only=True, frozen=True)
class Task(PulpObject):
    """Represents a Pulp task."""

    _SCHEMA = load_schema("task")

    id = pulp_attrib(type=str, pulp_field="task_id")
    """ID of this task (str)."""

    completed = pulp_attrib(default=None, type=bool)
    """True if this task has completed, successfully or otherwise.

    May be `None` if the state of this task is unknown.
    """

    succeeded = pulp_attrib(default=None, type=bool)
    """True if this task has completed successfully.

    May be `None` if the state of this task is unknown.
    """

    error_summary = pulp_attrib(default=None, type=str)
    """A summary of the reason for this task's failure (if any).

    This is a short string, generally a single line, suitable for display to users.
    The string includes the ID of the failed task.
    """

    error_details = pulp_attrib(default=None, type=str)
    """Detailed information for this task's failure (if any).

    This may be a multi-line string and may include technical information such as
    a Python backtrace generated by Pulp.

    ``error_details`` is a superset of the information available via ``error_summary``,
    so it is not necessary to display both.
    """

    tags = pulp_attrib(
        default=attr.Factory(frozenlist),
        type=list,
        converter=frozenlist,
        pulp_field="tags",
    )
    """The tags for this task.

    Typically includes info on the task's associated action and
    repo, such as:

    .. code-block:: python

        ["pulp:repository:rhel-7-server-rpms__7Server_x86_64",
         "pulp:action:publish"]
    """

    # TODO: is it a bug that this only allows a single repo ID??
    # Some tasks, like copy, involve multiple repos. We'll only include
    # the first repo ID from tags here... seems arbitrary.
    repo_id = pulp_attrib(type=str)
    """The ID of the repository associated with this task, otherwise None."""

    units = pulp_attrib(
        default=attr.Factory(frozenlist),
        type=list,
        pulp_field="result.units_successful",
        converter=frozenlist,
        pulp_py_converter=lambda raw: frozenlist(
            [Unit._from_task_data(x) for x in raw]
        ),
    )
    """Info on the units which were processed as part of this task
    (e.g. associated or unassociated).

    This is an iterable. Each element is an instance of
    :class:`~pubtools.pulplib.Unit` containing information on a processed
    unit.

    .. versionadded:: 1.5.0
    """

    units_data = pulp_attrib(
        default=attr.Factory(frozenlist),
        type=list,
        converter=frozenlist,
        pulp_field="result.units_successful",
    )
    """Info on the units which were processed as part of this task
    (e.g. associated or unassociated).

    This is a list. The list elements are the raw dicts as returned
    by Pulp. These should at least contain a 'type_id' and a 'unit_key'.

    .. deprecated:: 1.5.0
       Use :meth:`~pubtools.pulplib.Task.units` instead.
    """

    @repo_id.default
    def _repo_id_default(self):
        prefix = "pulp:repository:"
        for tag in self.tags or []:
            if tag.startswith(prefix):
                return tag[len(prefix) :]
        return None

    @succeeded.validator
    def _check_succeeded(self, _, value):
        if value and not self.completed:
            raise ValueError("Cannot have task with completed=False, succeeded=True")

    @classmethod
    def _data_to_init_args(cls, data):
        out = super(Task, cls)._data_to_init_args(data)

        state = data["state"]
        out["completed"] = state in ("finished", "error", "canceled", "skipped")
        out["succeeded"] = state in ("finished", "skipped")

        if state == "canceled":
            out["error_summary"] = "Pulp task [%s] was canceled" % data["task_id"]
            out["error_details"] = out["error_summary"]

        elif state == "error":
            out["error_summary"] = cls._error_summary(data)
            out["error_details"] = cls._error_details(data)

        return out

    @classmethod
    def _error_summary(cls, data):
        prefix = "Pulp task [%s] failed" % data["task_id"]
        error = data.get("error")
        if not error:
            return "%s: <unknown error>" % prefix
        return "%s: %s: %s" % (prefix, error["code"], error["description"])

    @classmethod
    def _error_details(cls, data):
        out = cls._error_summary(data)

        error = data.get("error")
        if not error:
            return out

        # Error looks like this:
        #
        # {
        #   'code': u'PLP0001',
        #   'data': {
        #     'message': 'a message'
        #   },
        #   'description': 'A general pulp exception occurred',
        #   'sub_errors': []
        # }
        #
        # See: https://docs.pulpproject.org/en/2.9/dev-guide/conventions/exceptions.html#error-details
        #
        # data can contain anything, or nothing.
        # It's only a convention that it often contains a message.
        #
        # sub_errors is currently ignored because I've never seen a non-empty
        # sub_errors yet.

        error_data = error.get("data") or {}
        messages = []

        # Message in a general exception
        if error_data.get("message"):
            messages.append(error_data["message"])

        # Some exceptions stash additional strings under details.errors
        if (error_data.get("details") or {}).get("errors"):
            error_messages = error_data["details"]["errors"]
            if isinstance(error_messages, list):
                messages.extend(error_messages)

        # Pulp docs refer to this as deprecated, but actually it's still
        # used and no alternative is provided.
        if data.get("traceback"):
            messages.append(data["traceback"])

        message = "\n".join(messages)
        if message:
            # message can have CRLF line endings in rare cases.
            message = message.replace("\r\n", "\n").strip()
            out = "%s:\n%s" % (out, _indent(message))

        return out


def _indent(text, level=2):
    spaces = " " * level
    return spaces + text.replace("\n", "\n" + spaces)



from frozenlist2 import frozenlist

from frozendict.core import frozendict  # pylint: disable=no-name-in-module
from ..compat_attr import Factory, validators, s, ib

optional = validators.optional
instance_of = validators.instance_of

optional_str = optional(instance_of(str))
optional_bool = optional(instance_of(bool))
optional_dict = optional(instance_of(dict))
optional_frozendict = optional(instance_of(frozendict))


@s(frozen=True, slots=True, hash=True)
class NamedMappingValidator:
    mapping = ib(
        validator=validators.deep_mapping(
            value_validator=validators.is_callable(),
            key_validator=validators.instance_of(str),
        ),
        default=Factory(dict),
    )
    type_ = ib(default=dict)

    def __call__(self, inst, attr, value):
        validators.instance_of(self.type_)(inst, attr, value)
        extra_values = set(value.keys() - self.mapping.keys())  # pylint: disable=E1101
        missing_values = set(
            self.mapping.keys() - value.keys()  # pylint: disable=E1101
        )  # pylint: disable=E1101
        if extra_values:
            raise ValueError(
                "Mapping {name} contains extra values {extra}".format(
                    name=attr.name, extra=extra_values
                )
            )
        if missing_values:
            raise ValueError(
                "Mapping {name} contains extra values {missing}".format(
                    name=attr.name, missing=missing_values
                )
            )

        for key, val in value.items():
            self.mapping[key](inst, attr, val)  # pylint: disable=E1136


def frozen_named_mapping_validator(mapping):
    return NamedMappingValidator(mapping, frozendict)


@s(kw_only=True, frozen=True, slots=True)
class ContainerListValidator(object):
    def __call__(self, inst, attr, value):
        validators.instance_of(frozenlist)(inst, attr, value)
        validators.deep_iterable(
            validators.and_(
                validators.instance_of(frozendict),
                validators.deep_mapping(
                    key_validator=validators.instance_of(str),
                    value_validator=frozen_named_mapping_validator(
                        {
                            "digest": validators.instance_of(str),
                            "images": validators.deep_mapping(
                                key_validator=validators.instance_of(str),
                                value_validator=validators.deep_mapping(
                                    key_validator=validators.instance_of(str),
                                    value_validator=validators.optional(
                                        validators.instance_of(str)
                                    ),
                                ),
                            ),
                        }
                    ),
                ),
            ),
            iterable_validator=validators.instance_of(frozenlist),
        )(inst, attr, value)


def container_list_validator():
    return ContainerListValidator()


def optional_list_of(elem_types):
    return optional(
        validators.deep_iterable(instance_of(elem_types), instance_of(list))
    )



import threading
import random
import uuid
import attr


from . import units

from pubtools.pulplib import FileUnit


class FakeState(object):
    # Private class holding all state associated with fake clients.
    #
    # A single state object can be accessed by multiple clients.
    # It can be thought of as similar to the (DB & filesystem) on a real
    # Pulp installation.
    #
    # Clients modifying state must explicitly lock this object.
    # Some read operations also require holding the lock.

    _DEFAULT_TYPE_IDS = [
        "distribution",
        "drpm",
        "erratum",
        "iso",
        "modulemd_defaults",
        "modulemd",
        "package_category",
        "package_environment",
        "package_group",
        "package_langpacks",
        "repository",
        "rpm",
        "srpm",
        "yum_repo_metadata_file",
    ]

    def __init__(self):
        self.repositories = []

        # Similar to a real Pulp server, units are stored through a "unit key"
        # layer of indirection:
        # - repos contain unit keys
        # - a unit key refers to one (and only one) unit
        #
        # By storing data in this way, we ensure we have similar behavior and
        # constraints as a real Pulp server, e.g. cannot store two different units
        # which would have the same key; updating a unit referenced from multiple repos
        # effectively updates it in all repos at once; etc.
        self.repo_unit_keys = {}
        self.units_by_key = {}

        self.publish_history = []
        self.upload_history = []
        self.uploads_pending = {}
        self.sync_history = []
        self.tasks = []
        self.maintenance_report = None
        self.type_ids = self._DEFAULT_TYPE_IDS[:]
        self.seen_unit_ids = set()
        self.lock = threading.Lock()
        self.uuidgen = random.Random()
        self.uuidgen.seed(0)
        self.unitmaker = units.UnitMaker(self.seen_unit_ids)
        # map of repo id => lock claims.
        self.repo_locks = {}
        # list containing lists of the lock state changes
        self.repo_lock_history = []

    def insert_repo_units(self, repo_id, units_to_add):
        # Insert an iterable of units into a specific repo.
        #
        # If a unit with the same key exists in multiple repos, this will update
        # matching units across *all* of those repos - same as a real pulp server.
        assert self.lock.locked()

        repo_unit_keys = self.repo_unit_keys.setdefault(repo_id, set())
        memberships = []
        if repo_id is not None:
            memberships.append(repo_id)

        for unit in units_to_add:
            # Always consume a unit ID from the unitmaker even if we don't actually
            # need it. The point of this is to ensure that if you serialize/deserialize
            # a fake client, e.g. as done in pubtools-pulp, then the freshly created
            # client will not try to use the same sequence of IDs as already used in
            # the units we've just deserialized.
            unit_id = self.unitmaker.next_unit_id()

            if not unit.unit_id:
                unit = attr.evolve(unit, unit_id=unit_id)

            self.seen_unit_ids.add(unit.unit_id)

            # Unit belongs to the repo we're adding it to.
            # Note: this may be further merged with an existing unit's
            # repository_memberships a few lines below.
            unit = attr.evolve(
                unit,
                repository_memberships=(unit.repository_memberships or [])
                + memberships,
            )

            if repo_id is not None:
                # Unit might be replacing earlier units in same repo.
                self.remove_clashing_units(repo_id, unit)

            unit_key = units.make_unit_key(unit)
            old_unit = self.units_by_key.get(unit_key)
            self.units_by_key[unit_key] = units.merge_units(old_unit, unit)
            repo_unit_keys.add(unit_key)

    def remove_clashing_units(self, repo_id, unit):
        assert self.lock.locked()

        # In preparation for adding 'unit' into a repo, remove any existing
        # units from that repo which would clash, in a compatible manner as Pulp.
        #
        # Currently this is a special case only for File units: although their
        # unit_key consists of more than just 'name' making it technically possible
        # to have multiple files with the same name in a repo, Pulp has special logic
        # to try to prevent this, so we do the same here.
        if isinstance(unit, FileUnit):
            for existing in self.repo_units(repo_id):
                if isinstance(existing, FileUnit) and existing.path == unit.path:
                    self.remove_unit(repo_id, existing)

    def remove_unit(self, repo_id, unit):
        # Remove a single unit from a repository.
        assert self.lock.locked()

        unit_key = units.make_unit_key(unit)

        # Ensure unit is no longer referenced from the repo
        self.repo_unit_keys[repo_id].remove(unit_key)

        # Ensure repo ID is no longer in memberships
        new_repos = [id for id in unit.repository_memberships if id != repo_id]
        self.units_by_key[unit_key] = attr.evolve(
            unit, repository_memberships=new_repos
        )

    @property
    def all_units(self):
        # Get all units in all repos.
        #
        # Cannot be used to modify units.
        return list(self.units_by_key.values())

    def repo_units(self, repo_id):
        # Get all units in a particular repo.
        #
        # Cannot be used to modify units.
        assert self.lock.locked()

        out = []
        for key in self.repo_unit_keys.get(repo_id) or []:
            out.append(self.units_by_key[key])
        return out

    def next_task_id(self):
        next_raw_id = self.uuidgen.randint(0, 2**128)
        return str(uuid.UUID(int=next_raw_id))

    def next_request_id(self):
        return self.next_task_id()



import re

from ..common import PulpObject
from ..attr import pulp_attrib, PULP2_FIELD
from ...util import dict_put
from ... import compat_attr as attr
from ...schema import load_schema


UNIT_CLASSES = {}


def unit_type(pulp_type):
    # decorator for Unit subclasses, registers against a
    # particular value of type_id.
    def decorate(klass):
        UNIT_CLASSES[pulp_type] = klass
        return klass

    return decorate


def type_ids_for_class(unit_class):
    # Given a concrete Unit subclass, returns those Pulp type id(s)
    # which may be used to find/load an object of that class.
    out = []

    for pulp_type, klass in UNIT_CLASSES.items():
        if klass is unit_class:
            out.append(pulp_type)

    return sorted(out)


def class_for_type_id(type_id):
    # Inverse of type_ids_for_class: given a Pulp type ID,
    # returns the corresponding Unit class, or None if there is no
    # corresponding class.
    return UNIT_CLASSES.get(type_id)


@attr.s(kw_only=True, frozen=True)
class Unit(PulpObject):
    """Represents a Pulp unit (a single piece of content).

    .. versionadded:: 1.5.0
    """

    _SCHEMA = load_schema("unit")

    content_type_id = pulp_attrib(type=str, pulp_field="_content_type_id")
    """The type of this unit.

    This value will match one of the content types returned by
    :meth:`~pubtools.pulplib.Client.get_content_type_ids`.
    """

    @classmethod
    def from_data(cls, data):
        # delegate to concrete subclass as needed
        if cls is Unit:
            type_id = data.get("_content_type_id")
            for klass_type_id, klass in UNIT_CLASSES.items():
                if klass_type_id == type_id:
                    return klass.from_data(data)

        return super(Unit, cls).from_data(data)

    @classmethod
    def _from_task_data(cls, data):
        # Like from_data, but massages the data from the format used in
        # task units_successful, which is slightly different from content search.
        unit_data = {}

        unit_data["_content_type_id"] = data.get("type_id")
        unit_data.update(data.get("unit_key") or {})

        return cls.from_data(unit_data)

    # A validator shared by various subclasses
    def _check_sum(self, value, sumtype, length):
        if value is None:
            return
        if not re.match(r"^[a-f0-9]{%s}$" % length, value):
            raise ValueError("Not a valid %s: %s" % (sumtype, value))

    @property
    def _usermeta(self):
        # Returns pulp_user_metadata dict for this unit.
        out = {}

        for field in self._usermeta_fields():
            pulp_field = field.metadata.get(PULP2_FIELD)
            python_value = getattr(self, field.name)
            pulp_value = PulpObject._any_to_data(python_value)
            dict_put(out, pulp_field, pulp_value)

        return out.get("pulp_user_metadata") or {}

    @classmethod
    def _usermeta_fields(cls):
        # Returns the subset of fields on this class which are stored under the
        # pulp_user_metadata dict. In public API we refer to these as 'mutable' fields.
        return [
            fld
            for fld in attr.fields(cls)
            if fld.metadata.get(PULP2_FIELD).startswith("pulp_user_metadata.")
        ]

    @classmethod
    def _usermeta_from_kwargs(cls, **kwargs):
        # Given kwargs mapping to mutable fields on this unit class, returns a dict
        # of the form:
        #
        #   {"pulp_user_metadata": {...serialized mutable fields}}
        #
        # ...suitable for merging into a 'metadata' dict on a content upload.
        #
        # If any of the kwargs do not map to a mutable field, an exception is raised.
        #
        fields = cls._usermeta_fields()
        out = {}

        for field in fields:
            if not field.name in kwargs:
                continue

            pulp_field = field.metadata.get(PULP2_FIELD)
            python_value = kwargs.pop(field.name)
            pulp_value = PulpObject._any_to_data(python_value)
            dict_put(out, pulp_field, pulp_value)

        # Ensure that we have consumed all arguments; if not, that means the
        # caller tried to set something we don't support.
        remaining = sorted(kwargs.keys())
        if remaining:
            raise ValueError(
                "Not mutable %s field(s): %s" % (cls.__name__, ", ".join(remaining))
            )

        return out



import re

from frozenlist2 import frozenlist

from more_executors.futures import f_map, f_proxy, f_return, f_zip, f_flat_map
from .base import Repository, SyncOptions, repo_type, Importer
from ..attr import pulp_attrib
from ..common import DetachedException
from ...model.unit import RpmUnit
from ... import compat_attr as attr, comps
from ...criteria import Criteria, Matcher


@attr.s(kw_only=True, frozen=True)
class YumImporter(Importer):
    type_id = pulp_attrib(
        default="yum_importer", type=str, pulp_field="importer_type_id"
    )
    """
    Specific importer_type_id for Yum repositories.
    """


@attr.s(kw_only=True, frozen=True)
class YumSyncOptions(SyncOptions):
    """Options controlling a yum repository
    :meth:`~pubtools.pulplib.Repository.sync`.
    """

    query_auth_token = pulp_attrib(default=None, type=str)
    """An authorization token that will be added to every request made to the feed URL's server
    """

    max_downloads = pulp_attrib(default=None, type=int)
    """Number of threads used when synchronizing the repository.
    """

    remove_missing = pulp_attrib(default=None, type=bool)
    """If true, as the repository is synchronized, old rpms will be removed.
    """

    retain_old_count = pulp_attrib(default=None, type=int)
    """Count indicating how many old rpm versions to retain.
    """

    skip = pulp_attrib(default=None, type=list)
    """List of content types to be skipped during the repository synchronization
    """

    checksum_type = pulp_attrib(default=None, type=str)
    """checksum type to use for metadata generation.

    Defaults to source checksum type of sha256
    """

    num_retries = pulp_attrib(default=None, type=int)
    """Number of times to retry before declaring an error during repository synchronization

    Default is 2.
    """

    download_policy = pulp_attrib(default=None, type=str)
    """Set the download policy for a repository.

    Supported options are immediate,on_demand,background
    """

    force_full = pulp_attrib(default=None, type=bool)
    """Boolean flag. If true, full re-sync is triggered.
    """

    require_signature = pulp_attrib(default=None, type=bool)
    """Requires that imported packages like RPM/DRPM/SRPM should be signed
    """

    allowed_keys = pulp_attrib(default=None, type=list)
    """List of allowed signature key IDs that imported packages can be signed with
    """


@repo_type("rpm-repo")
@attr.s(kw_only=True, frozen=True)
class YumRepository(Repository):
    """A :class:`~pubtools.pulplib.Repository` for RPMs, errata and related content."""

    # this class only overrides some defaults for attributes defined in super

    type = pulp_attrib(default="rpm-repo", type=str, pulp_field="notes._repo-type")

    population_sources = pulp_attrib(
        default=attr.Factory(frozenlist),
        type=list,
        converter=frozenlist,
        pulp_field="notes.population_sources",
        mutable=True,
    )
    """List of repository IDs used to populate this repository
    """

    ubi_population = pulp_attrib(
        default=False,
        type=bool,
        pulp_field="notes.ubi_population",
        mutable=True,
    )
    """Flag indicating whether repo should be populated from population_sources for the purposes of UBI
    """

    mutable_urls = attr.ib(
        default=attr.Factory(lambda: frozenlist(["repodata/repomd.xml"])),
        type=list,
        converter=frozenlist,
    )

    ubi_config_version = pulp_attrib(
        default="",
        type=str,
        pulp_field="notes.ubi_config_version",
        mutable=True,
    )
    """Version of UBI config that should be used for population of this repository."""

    importer = pulp_attrib(
        default=YumImporter(),
        type=YumImporter,
        pulp_field="importers",
        pulp_py_converter=YumImporter._from_data,
        py_pulp_converter=YumImporter._to_data,
    )
    """
    An object of :class:`~pubtools.pulplib.YumImporter` that is associated with the repository.

    .. versionadded:: 2.39.0
    """

    def get_binary_repository(self):
        """Find and return the binary repository relating to this repository.

        Yum repositories usually come in triplets of
        (binary RPMs, debuginfo RPMs, source RPMs). For example:

        .. list-table::
            :widths: 75 25

            * - ``rhel-7-server-rpms__7Server__x86_64``
              - binary
            * - ``rhel-7-server-debug-rpms__7Server__x86_64``
              - debug
            * - ``rhel-7-server-source-rpms__7Server__x86_64``
              - source

        This method along with :meth:`get_debug_repository` and :meth:`get_source_repository` allow locating other repositories
        from within this group.

        Returns:
            ``Future[YumRepository]``
                Binary repository relating to this repository.
            ``Future[None]``
                If there is no related repository.
        """
        return self._get_related_repository(repo_t="binary")

    def get_debug_repository(self):
        """Find and return the debug repository relating to this repository.

        Returns:
            ``Future[YumRepository]``
                Debug repository relating to this repository.
            ``Future[None]``
                If there is no related repository.
        """
        return self._get_related_repository(repo_t="debug")

    def get_source_repository(self):
        """Find and return the source repository relating to this repository.

        Returns:
            ``Future[YumRepository]``
                Source repository relating to this repository.
            ``Future[None]``
                If there is no related repository.
        """
        return self._get_related_repository(repo_t="source")

    def _get_related_repository(self, repo_t):
        if not self._client:
            raise DetachedException()

        suffixes_mapping = {
            "binary": "/os",
            "debug": "/debug",
            "source": "/source/SRPMS",
        }

        regex = r"(/os|/source/SRPMS|/debug)$"

        def unpack_page(page):
            if len(page.data) != 1:
                return None

            return page.data[0]

        suffix = suffixes_mapping[repo_t]
        if str(self.relative_url).endswith(suffix):
            return f_proxy(f_return(self))

        base_url = re.sub(regex, "", self.relative_url)
        relative_url = base_url + suffix
        criteria = Criteria.and_(
            Criteria.with_field("relative_url", relative_url),
            Criteria.with_field("type_id", "yum_distributor"),
        )
        page_f = self._client.search_distributor(criteria)
        distributor_f = f_proxy(f_map(page_f, unpack_page))

        if distributor_f.result() is None:
            out = f_return()
        else:
            out = self._client.get_repository(distributor_f.repo_id)
        return out

    def upload_rpm(self, file_obj, **kwargs):
        """Upload an RPM to this repository.

        .. warning::

            For RPMs belonging to a module, it's strongly advised to upload
            the module metadata first (using :meth:`upload_modules`) and only
            proceed with uploading RPMs once module upload has completed.

            This reduces the risk of accidentally publishing a repository with
            modular RPMs without the corresponding metadata (which has a much
            worse impact than publishing metadata without the corresponding RPMs).

        Args:
            file_obj (str, file object)
                If it's a string, then it's the path of an RPM to upload.

                Otherwise, it should be a
                `file-like object <https://docs.python.org/3/glossary.html#term-file-object>`_
                pointing at the bytes to upload.
                The client takes ownership of this file object; it should
                not be modified elsewhere, and will be closed when upload
                completes.

            kwargs
                Additional field values to set on the uploaded unit.

                Any :class:`~pubtools.pulplib.RpmUnit` fields documented as
                *mutable* may be included here (for example, ``cdn_path``).
                An error will occur if attempting to set other fields.

        Returns:
            Future[list of :class:`~pubtools.pulplib.Task`]
                A future which is resolved after content has been imported
                to this repo.

        Raises:
            DetachedException
                If this instance is not attached to a Pulp client.

        .. versionadded:: 2.16.0

        .. versionadded:: 2.20.0
            Added ability to set mutable fields on upload.
        """
        # We want some name of what we're uploading for logging purposes, but the
        # input could be a plain string, or a file object with 'name' attribute, or
        # a file object without 'name' ... make sure we do something reasonable in
        # all cases.
        if isinstance(file_obj, str):
            name = file_obj
        else:
            # If we don't know what we're uploading we just say it's "an RPM"...
            name = getattr(file_obj, "name", "an RPM")

        unit_metadata_fn = None

        usermeta = RpmUnit._usermeta_from_kwargs(**kwargs)
        if usermeta:
            unit_metadata_fn = lambda _: usermeta

        return self._upload_then_import(
            file_obj, name, "rpm", unit_metadata_fn=unit_metadata_fn
        )

    def upload_metadata(self, file_obj, metadata_type):
        """Upload a metadata file to this repository.

        A metadata file is any additional file which will be published alongside,
        and referenced from, the repodata ``.xml`` and ``.sqlite`` files when this
        repo is published.

        Args:
            file_obj (str, file object)
                If it's a string, then it's the path of a file to upload.

                Otherwise, it should be a
                `file-like object <https://docs.python.org/3/glossary.html#term-file-object>`_
                pointing at the bytes to upload.
                The client takes ownership of this file object; it should
                not be modified elsewhere, and will be closed when upload
                completes.

            metadata_type (str)
                Identifies the type of metadata being uploaded.

                This is an arbitrary string which will be reproduced in the yum
                repo metadata on publish. The appropriate value depends on the
                type of data being uploaded. For example, ``"productid"`` should
                be used when uploading an RHSM-style product certificate.

                A repository may only contain a single metadata file of each type.
                If a file of this type is already present in the repo, it will be
                overwritten by the upload.

        Returns:
            Future[list of :class:`~pubtools.pulplib.Task`]
                A future which is resolved after content has been imported
                to this repo.

        Raises:
            DetachedException
                If this instance is not attached to a Pulp client.

        .. versionadded:: 2.17.0
        """
        if isinstance(file_obj, str):
            name = "%s (%s)" % (file_obj, metadata_type)
        else:
            # If we don't know what we're uploading we just say "<type> metadata"...
            name = getattr(file_obj, "name", "%s metadata" % metadata_type)

        return self._upload_then_import(
            file_obj,
            name,
            "yum_repo_metadata_file",
            # Requirements around unit key and metadata can be found at:
            # https://github.com/pulp/pulp_rpm/blob/5c5a7dcc058b29d89b3a913d29cfcab41db96686/plugins/pulp_rpm/plugins/importers/yum/upload.py#L246
            unit_key_fn=lambda _: {"data_type": metadata_type, "repo_id": self.id},
            unit_metadata_fn=lambda upload: {
                "checksum": upload[0],
                "checksum_type": "sha256",
            },
        )

    def upload_modules(self, file_obj):
        """Upload a modulemd stream to this repository.

        All supported documents in the given stream will be imported to this
        repository. On current versions of Pulp 2.x, this means only:

        * `modulemd v2 <https://github.com/fedora-modularity/libmodulemd/blob/main/yaml_specs/modulemd_stream_v2.yaml>`_
        * `modulemd-defaults v1 <https://github.com/fedora-modularity/libmodulemd/blob/main/yaml_specs/modulemd_defaults_v1.yaml>`_

        Attempting to use other document types may result in an error.

        Args:
            file_obj (str, file object)
                If it's a string, then it's the path of a modulemd YAML
                file to upload.

                Otherwise, it should be a
                `file-like object <https://docs.python.org/3/glossary.html#term-file-object>`_
                pointing at the text to upload.
                The client takes ownership of this file object; it should
                not be modified elsewhere, and will be closed when upload
                completes.

        Returns:
            Future[list of :class:`~pubtools.pulplib.Task`]
                A future which is resolved after content has been imported
                to this repo.

        Raises:
            DetachedException
                If this instance is not attached to a Pulp client.

        .. versionadded:: 2.17.0
        """
        if isinstance(file_obj, str):
            name = file_obj
        else:
            name = getattr(file_obj, "name", "modulemds")

        return self._upload_then_import(file_obj, name, "modulemd")

    def upload_comps_xml(self, file_obj):
        """Upload a comps.xml file to this repository.

        .. warning::

            Beware of the following quirks with respect to the upload of comps.xml:

            * Pulp does not directly store the uploaded XML. Instead, this library
              parses the XML and uses the content to store various units. The comps
              XML rendered as a yum repository is published is therefore not
              guaranteed to be bytewise-identical to the uploaded content.

            * The uploaded XML must contain all comps data for the repo, as
              any existing comps data will be removed from the repo.

            * The XML parser is not secure against maliciously constructed data.

            * The process of parsing the XML and storing units consists of multiple
              steps which cannot be executed atomically. That means *if this
              operation is interrupted, the repository may be left with incomplete
              data*. It's recommended to avoid publishing a repository in this state.

        Args:
            file_obj (str, file object)
                If it's a string, then it's the path of a comps XML
                file to upload.

                Otherwise, it should be a
                `file-like object <https://docs.python.org/3/glossary.html#term-file-object>`_
                pointing at the bytes of a valid comps.xml file.

                The client takes ownership of this file object; it should
                not be modified elsewhere, and will be closed when upload
                completes.

        Returns:
            Future[list of :class:`~pubtools.pulplib.Task`]
                A future which is resolved after content has been imported
                to this repo.

        Raises:
            DetachedException
                If this instance is not attached to a Pulp client.

        .. versionadded:: 2.17.0
        """
        if isinstance(file_obj, str):
            file_name = file_obj
            file_obj = open(file_obj, "rb")
        else:
            file_name = getattr(file_obj, "name", "comps.xml")

        # Parse the provided XML. We will crash here if the given XML is not
        # valid.
        with file_obj:
            unit_dicts = comps.units_for_xml(file_obj)

        # Every comps-related unit type has a repo_id which should reference the repo
        # we're uploading to.
        for unit in unit_dicts:
            unit["repo_id"] = self.id

        comps_type_ids = [
            "package_group",
            "package_category",
            "package_environment",
            "package_langpacks",
        ]

        # Remove former units of comps-related types so that the end result is only
        # those units included in the current XML.
        crit = Criteria.with_field("content_type_id", Matcher.in_(comps_type_ids))
        out = self.remove_content(crit)

        # Once removal is done we can upload each unit.
        upload_f = []
        for unit_dict in unit_dicts:
            type_id = unit_dict["_content_type_id"]

            # For one comps.xml we are doing multiple upload operations, each of
            # which would be logged independently. Come up with some reasonable name
            # for each unit to put into the logs.
            #
            # Example: if uploading my-comps.xml and processing a package_group
            # with id kde-desktop-environment, the name for logging purposes would
            # be: "my-comps.xml [group.kde-desktop-environment]".
            #
            unit_name = type_id.replace("package_", "")
            if unit_dict.get("id"):
                unit_name = "%s.%s" % (unit_name, unit_dict["id"])
            unit_name = "%s [%s]" % (file_name, unit_name)

            upload_f.append(
                f_flat_map(
                    out, self._comps_unit_uploader(unit_name, type_id, unit_dict)
                )
            )

        # If there were no units to upload then just return the removal.
        if not upload_f:
            return out

        # There were uploads, then we'll wait for all of them to complete and
        # return the tasks for all.
        out = f_zip(*upload_f)
        out = f_map(out, lambda uploads: sum(uploads, []))

        return out

    def _comps_unit_uploader(self, name, type_id, metadata):
        # A helper used from upload_comps_xml.
        #
        # This helper only exists to eagerly bind arguments for a single
        # unit upload, due to confusing behavior around variable scope
        # when combining loops and lambdas.

        def upload(_unused):
            return self._upload_then_import(
                file_obj=None,
                name=name,
                type_id=type_id,
                unit_metadata_fn=lambda _: metadata,
            )

        return upload

    def upload_erratum(self, erratum):
        """Upload an erratum/advisory object to this repository.

        .. warning::

            There are many quirks with respect to advisory upload. Please be aware
            of the following before using this API:

            * Only one advisory with a given ``id`` may exist in the system.

            * When uploading an advisory with an ``id`` equal to one already in the
              system, the upload will generally be ignored (i.e. complete successfully
              but have no effect), unless either the ``version`` or ``updated`` fields
              have a value larger than the existing advisory.

              This implies that, if you want to ensure an existing advisory is updated,
              you must first search for the existing object and mutate one of these
              fields before uploading a modified object. *The library will not take
              care of this for you.*

            * When overwriting an existing advisory, all fields will be overwritten.
              The sole exception is the ``pkglist`` field which will be merged with
              existing data when applicable.

            * If an advisory with the same ``id`` is present in multiple published yum
              repositories with inconsistent fields, yum/dnf client errors or warnings
              may occur. It's therefore recommended that, whenever an existing
              advisory is modified, every repository containing that advisory should
              be republished. *The library will not take care of this for you.*

            * The ``repository_memberships`` field on the provided object has no effect
              (it cannot be used to upload an advisory to multiple repos at once).

        Args:
            erratum (:class:`~pubtools.pulplib.ErratumUnit`)
                An erratum object.

                Unlike most other uploaded content, errata are not backed by any
                file; any arbitrarily constructed ErratumUnit may be uploaded.

        Returns:
            Future[list of :class:`~pubtools.pulplib.Task`]
                A future which is resolved after content has been imported
                to this repo.

        Raises:
            DetachedException
                If this instance is not attached to a Pulp client.

        .. versionadded:: 2.17.0
        """
        # Convert from ErratumUnit to a raw Pulp-style dict, recursively.
        erratum_dict = erratum._to_data()

        # Drop this one field because the _content_type_id, though embedded
        # in unit dicts on read, is passed as a separate parameter on write.
        type_id = erratum_dict.pop("_content_type_id")

        # Drop this because the caller cannot influence the _id (unit id)
        # for uploaded units.
        del erratum_dict["_id"]

        # And drop this one because repository_memberships is synthesized when
        # Pulp renders units, and can't be set during import.
        del erratum_dict["repository_memberships"]

        return self._upload_then_import(
            file_obj=None,
            name=erratum_dict["id"],
            type_id=type_id,
            unit_key_fn=lambda _: {"id": erratum_dict["id"]},
            unit_metadata_fn=lambda _: erratum_dict,
        )



from ._impl.client import Client, PulpException, TaskFailedException, CopyOptions
from ._impl.criteria import Criteria, Matcher
from ._impl.page import Page
from ._impl.model import (
    PulpObject,
    DetachedException,
    InvalidDataException,
    Repository,
    YumRepository,
    FileRepository,
    ContainerImageRepository,
    Unit,
    FileUnit,
    RpmUnit,
    RpmDependency,
    ErratumUnit,
    ErratumReference,
    ErratumPackageCollection,
    ErratumPackage,
    ErratumModule,
    ModulemdUnit,
    ModulemdDependency,
    ModulemdDefaultsUnit,
    YumRepoMetadataFileUnit,
    Distributor,
    PublishOptions,
    FileSyncOptions,
    ContainerSyncOptions,
    YumSyncOptions,
    SyncOptions,
    Task,
    MaintenanceReport,
    MaintenanceEntry,
    Importer,
    YumImporter,
    FileImporter,
)
from ._impl.fake import FakeController



# -*- coding: utf-8 -*-

from xml.parsers import expat

from io import StringIO


# All optional fields in the pulp-rpm comps.xml units
COMMON_FIELDS = [
    "description",
    "translated_description",
    "display_order",
    "translated_name",
    "name",
]
UNIT_FIELD_DEFAULTS = {
    "package_group": (
        {"default": False, "user_visible": False},
        [
            "default_package_names",
            "optional_package_names",
            "mandatory_package_names",
            "default",
            "user_visible",
            "langonly",
            "conditional_package_names",
        ]
        + COMMON_FIELDS,
    ),
    "package_category": ({}, ["packagegroupids"] + COMMON_FIELDS),
    "package_environment": ({}, ["group_ids", "options"] + COMMON_FIELDS),
    "package_langpacks": ({}, ["matches"]),
}


class BooleanStringIO(StringIO):
    """A StringIO which coerces the output value into a boolean."""

    def getvalue(self):
        # Note: StringIO in python2 is an old-style class, so no super().
        value = StringIO.getvalue(self).strip()
        if value.lower() in ("false", ""):
            return False
        return True


class IntegerStringIO(StringIO):
    """A StringIO which coerces the output value into an integer."""

    def getvalue(self):
        value = StringIO.getvalue(self).strip()
        if not value:
            return None
        return int(value)


def close_buffers(value):
    """Returns a copy of 'value' with all StringIO buffers replaced with the
    content of each buffer.

    This is used at the end of parsing, once we know that no more text data can
    arrive, to seal the output and convert into a serializable form.
    """

    if isinstance(value, list):
        return [close_buffers(elem) for elem in value]

    if isinstance(value, dict):
        out = {}
        for (key, elem) in value.items():
            out[key] = close_buffers(elem)
        return out

    if isinstance(value, StringIO):
        return value.getvalue()

    return value


class CompsParser(object):
    """A stateful parser for comps.xml data.

    This parser wraps xml.parsers.expat and installs handlers which are able to
    load XML elements into the form used by Pulp for the relevant unit types.
    """

    def __init__(self):
        self.raw_parser = expat.ParserCreate()

        # Current parse state. The output is ultimately returned from 'units'.
        self.units = []
        self.current_buf = StringIO()
        self.current_unit = {}
        self.current_path = []

        # Bind ourselves to expat.
        for name in dir(self):
            if name.endswith("Handler"):
                setattr(self.raw_parser, name, getattr(self, name))

    def parse(self, io):
        """Parse comps XML from io, a file-like object in binary mode.

        Returns parsed units in Pulp form (i.e. a list of dicts with each dict
        having the Pulp-specific attributes such as _content_type_id, ...)
        """
        self.units = []
        self.current_unit = {}
        self.current_path = []
        self.raw_parser.ParseFile(io)
        self.units = close_buffers(self.units)
        return self.units

    ########################## Shared handlers #########################
    # Handle some basic tags shared by multiple types such as:
    #
    #     <id>kde-desktop-environment</id>
    #     <name>KDE Desktop</name>
    #     <name xml:lang="af">KDE-werkskerm</name>
    #     <name xml:lang="as">KDE ডেস্কটপ</name>
    #     ...
    #     <description>The KDE SC includes ...</description>
    #     <description xml:lang="ar">الـ KDE SC يتضمن سطح...</description>

    def handle_text_elem(self, tag, attrs):
        if tag in ("name", "description") and attrs.get("xml:lang"):
            # Elem like this: <name xml:lang="af">3D-drukwerk</name>
            # Is parsed into: translated_name["af"] = "3D-drukwerk"
            tag = "translated_" + tag
            lang = attrs["xml:lang"]
            self.current_buf = StringIO()
            self.current_unit.setdefault(tag, {})[lang] = self.current_buf
            return True

        elif tag in ("id", "name", "description"):
            self.current_buf = StringIO()
            self.current_unit[tag] = self.current_buf
            return True

        return False

    def handle_display_order_tag(self, tag):
        if tag == "display_order":
            self.current_buf = IntegerStringIO()
            self.current_unit[tag] = self.current_buf
            return True
        return False

    ########################## Group ###################################
    #   <group>
    #     <id>3d-printing</id>
    #     <name>3D Printing</name>
    #     <name xml:lang="af">3D-drukwerk</name>
    #     <name xml:lang="bg">3D Печатане</name>
    #     ...
    #     <description>3D printing software</description>
    #     <description xml:lang="af">3D-druksagteware</description>
    #     <description xml:lang="bg">Софтуер за 3D печатане</description>
    #     ...
    #     <default>false</default>
    #     <uservisible>true</uservisible>
    #     <packagelist>
    #       <packagereq type="default">admesh</packagereq>
    #       <packagereq type="default">blender</packagereq>
    #       ...
    #     </packagelist>
    #   </group>

    def start_group_elem(self):
        self.current_unit = {"_content_type_id": "package_group"}
        self.units.append(self.current_unit)

    def handle_group_tag(self, tag, attrs):
        if self.handle_text_elem(tag, attrs):
            return

        elif tag in ("default", "uservisible"):
            self.current_buf = BooleanStringIO()
            if tag == "uservisible":
                tag = "user_visible"
            self.current_unit[tag] = self.current_buf

    def handle_group_packagelist(self, attrs):
        package_type = attrs.get("type") or "mandatory"
        key = package_type + "_package_names"
        self.current_buf = StringIO()
        target = self.current_unit.setdefault(key, [])

        if package_type == "conditional":
            # "conditional" type is special, a requires attrib is included, and Pulp
            # stores a (pkgname, requires) tuple.
            target.append([self.current_buf, attrs.get("requires")])
        else:
            # Anything else just stores the package name.
            target.append(self.current_buf)

    def handle_group_elem(self, path, attrs):
        if path == []:
            return self.start_group_elem()

        if len(path) == 1:
            return self.handle_group_tag(path[0], attrs)

        if path == ["packagelist", "packagereq"]:
            return self.handle_group_packagelist(attrs)

    ########################## Environment ################################
    #   <environment>
    #     <id>cloud-server-environment</id>
    #     <name>Fedora Cloud Server</name>
    #     <name xml:lang="af">Fedora-wolkbediener</name>
    #     <name xml:lang="bg">Fedora Cloud Сървър</name>
    #     ...
    #     <description>A server install with components needed...</description>
    #     <description xml:lang="af">’n Bedienerinstallasie met die nodige...</description>
    #     ...
    #     <display_order>3</display_order>
    #     <grouplist>
    #       <groupid>cloud-server</groupid>
    #       <groupid>core</groupid>
    #     </grouplist>
    #     <optionlist>
    #       <groupid>directory-server</groupid>
    #       <groupid>dns-server</groupid>
    #       ...
    #     </optionlist>
    #   </environment>

    def start_environment_elem(self):
        self.current_unit = {"_content_type_id": "package_environment"}
        self.units.append(self.current_unit)

    def handle_environment_tag(self, tag, attrs):
        if self.handle_text_elem(tag, attrs):
            return

        self.handle_display_order_tag(tag)

    def handle_environment_grouplist(self):
        self.current_buf = StringIO()
        self.current_unit.setdefault("group_ids", []).append(self.current_buf)

    def handle_environment_optionlist(self, attrs):
        self.current_buf = StringIO()

        option = {"group": self.current_buf}

        option["default"] = (attrs.get("default") or "").lower() == "true"

        self.current_unit.setdefault("options", []).append(option)

    def handle_environment_elem(self, path, attrs):
        if path == []:
            return self.start_environment_elem()

        if len(path) == 1:
            return self.handle_environment_tag(path[0], attrs)

        if path == ["grouplist", "groupid"]:
            return self.handle_environment_grouplist()

        if path == ["optionlist", "groupid"]:
            return self.handle_environment_optionlist(attrs)

    ########################## Category ###################################
    #   <category>
    #     <id>kde-desktop-environment</id>
    #     <name>KDE Desktop</name>
    #     <name xml:lang="af">KDE-werkskerm</name>
    #     <name xml:lang="as">KDE ডেস্কটপ</name>
    #     ...
    #     <description>The KDE SC includes ...</description>
    #     <description xml:lang="ar">الـ KDE SC يتضمن سطح...</description>
    #     <display_order>10</display_order>
    #     <grouplist>
    #       <groupid>kde-apps</groupid>
    #       <groupid>kde-desktop</groupid>
    #       ...
    #     </grouplist>
    #   </category>

    def start_category_elem(self):
        self.current_unit = {"_content_type_id": "package_category"}
        self.units.append(self.current_unit)

    def handle_category_tag(self, tag, attrs):
        if self.handle_text_elem(tag, attrs):
            return

        self.handle_display_order_tag(tag)

    def handle_category_grouplist(self):
        self.current_buf = StringIO()
        self.current_unit.setdefault("packagegroupids", []).append(self.current_buf)

    def handle_category_elem(self, path, attrs):
        if path == []:
            return self.start_category_elem()

        if len(path) == 1:
            return self.handle_category_tag(path[0], attrs)

        if path == ["grouplist", "groupid"]:
            return self.handle_category_grouplist()

    ####################### Langpacks #####################################
    # <langpacks>
    #     <match install="LabPlot-doc-%s" name="LabPlot-doc"/>
    #     <match install="aspell-%s" name="aspell"/>
    #     ...
    # </langpacks>

    def start_langpacks_elem(self):
        self.current_unit = {"_content_type_id": "package_langpacks"}
        self.units.append(self.current_unit)

    def handle_langpacks_match(self, attrs):
        self.current_unit.setdefault("matches", []).append(
            {"install": attrs.get("install"), "name": attrs.get("name")}
        )

    def handle_langpacks_elem(self, path, attrs):
        if path == []:
            return self.start_langpacks_elem()

        if path == ["match"]:
            return self.handle_langpacks_match(attrs)

    ####################### Handlers ######################################
    # All of these are installed on the expat parser.
    #
    # The name of each method must exactly match the names documented at
    # xml.parsers.expat.
    #
    # These handlers dispatch to the rest of the code above to parse elements
    # once the type is known.

    def StartElementHandler(self, name, attrs):
        self.current_path.append(name)

        prefix = self.current_path[:2]
        rest = self.current_path[2:]

        if prefix == ["comps", "group"]:
            return self.handle_group_elem(rest, attrs)

        if prefix == ["comps", "category"]:
            return self.handle_category_elem(rest, attrs)

        if prefix == ["comps", "environment"]:
            return self.handle_environment_elem(rest, attrs)

        if prefix == ["comps", "langpacks"]:
            return self.handle_langpacks_elem(rest, attrs)

    def CharacterDataHandler(self, data):
        self.current_buf.write(data)

    def EndElementHandler(self, _name):
        self.current_path = self.current_path[:-1]
        self.current_buf = StringIO()


def units_for_xml(io):
    """Parse comps.xml from a file-like object and return corresponding Pulp units.

    Arguments:
        io (file object)
            A file-like object.

            Should be opened in binary mode, and should point at bytes making up
            a valid comps.xml document encoded with UTF-8, UTF-16, ISO-8859-1 (Latin1)
            or ASCII.

    Returns:
        list[dict]
            A list of unit metadata in the form used by Pulp2 for unit types relating
            to comps.xml (e.g. package_group, package_langpacks, ...).

            Returned units omit the 'repo_id' attribute, which must be filled in if
            units will be uploaded to a specific repo.

    Note: this is the only function in this module intended to be used from other
    modules.
    """
    parser = CompsParser()
    return fill_unit_field_defaults(parser.parse(io))


def fill_unit_field_defaults(units):
    """
    Fill in missing comp unit fields with default values.

    Arguments:
    units (dict)
        Parsed comps.xml file.
    """

    for unit in units:
        defaults, fields = UNIT_FIELD_DEFAULTS[unit["_content_type_id"]]
        for field in fields:
            unit[field] = unit.get(field, defaults.get(field))

    return units



import datetime
import functools

from frozenlist2 import frozenlist
from frozendict.core import frozendict  # pylint: disable=no-name-in-module

from .attr import PULP2_PY_CONVERTER


def get_converter(field, value):
    """Given an attrs target field and an input value from Pulp,
    return a converter function which should be used to convert the Pulp value
    into a Python representation."""

    metadata_converter = field.metadata.get(PULP2_PY_CONVERTER)
    if metadata_converter:
        # explicitly defined for this field, just return it
        return metadata_converter

    # Nothing explicitly defined, but check the types, there may still be
    # some applicable default
    if field.type is datetime.datetime and isinstance(value, str):
        return read_timestamp

    return null_convert


def null_convert(value):
    return value


def read_timestamp(value):
    try:
        return datetime.datetime.strptime(value, "%Y-%m-%dT%H:%M:%SZ")
    except ValueError:
        # irritatingly (probably a bug), some values were missing the "Z"
        # technically meaning we don't know the timezone.
        # So we try parsing again without it and we just assume it's UTC.
        return datetime.datetime.strptime(value, "%Y-%m-%dT%H:%M:%S")


def tolerant_timestamp(value):
    # Converter for fields which can accept a timestamp string, but which
    # falls back to returning the input verbatim if conversion fails.
    #
    # Since it tolerates failed conversions, this is intended to be combined
    # with a validator.
    if isinstance(value, str):
        try:
            return read_timestamp(value)
        except ValueError:
            # Not a timestamp, conversion doesn't happen
            pass

    return value


def timestamp_converter(value):
    # Converter for fields which are stored as strings,
    # but which model is expecting datetime
    # falls back to returning the input verbatim if not a datetime.
    #
    if isinstance(value, datetime.datetime):
        return value.strftime("%Y-%m-%dT%H:%M:%SZ")

    return value


def write_timestamp(value):
    # defaults to current time if value is None
    if value is None:
        value = datetime.datetime.utcnow()
    return value.strftime("%Y-%m-%dT%H:%M:%SZ")


def frozenlist_or_none_converter(obj, map_fn=(lambda x: x)):
    if obj is not None:
        return frozenlist(map_fn(obj))
    return None


frozenlist_or_none_sorted_converter = functools.partial(
    frozenlist_or_none_converter, map_fn=lambda x: sorted(set(x))
)


def frozendict_or_none_converter(obj):
    # Convert object and values to immutable structures.
    # Skip convert if the obj is already frozendict.
    # This happens when the class containing the value is copied (e.g: with attr.evolve).
    if obj is not None and not isinstance(obj, frozendict):
        for (key, value) in obj.items():
            if isinstance(value, list):
                obj[key] = frozenlist_or_none_converter(value)
            if isinstance(value, dict):
                obj[key] = frozendict_or_none_converter(value)
        return frozendict(obj)
    return obj


def freeze(obj):
    """Convert complex object composed of dicts and lists to equivalent object composed of
    frozendicts and frozenlists.
    """

    ret = [None]
    stack = [(obj, ret, 0)]
    traversal = []

    # interate over the structure and do post-order traversal
    while stack:
        cobj, cparent, ckey = stack.pop(0)

        # cobj_replacement represents unfrozen object. This way it's possible to freeze
        # already frozen or partially frozen structure
        cobj_replacement = cobj
        if isinstance(cobj, frozenlist):
            cobj_replacement = [None] * len(cobj)
        if isinstance(cobj, frozendict):
            cobj_replacement = {}

        if isinstance(cobj_replacement, list):
            for nth, i in enumerate(cobj):
                stack.insert(0, (i, cobj_replacement, nth))
        elif isinstance(cobj_replacement, dict):
            for key in cobj:
                stack.insert(0, (cobj[key], cobj_replacement, key))
        traversal.insert(0, (cobj_replacement, cparent, ckey))

    # Walk through traversal record. As traversal is post-order
    # leafs are processed first and therefore nested dict/lists
    # are replaced froze first
    for titem in traversal:
        cobj, cparent, ckey = titem
        # if traversal entry is list replace it with frozenlist
        if isinstance(cobj, list):
            cparent[ckey] = frozenlist(cobj)
        # if traversal entry is list replace it with frozendict
        elif isinstance(cobj, dict):
            cparent[ckey] = frozendict(cobj)
        else:
            cparent[ckey] = cobj

    # return firt items of last object in traversal, which is ret[0].
    return traversal[-1][1][0]


def freeze_or_empty(obj):
    if obj is None:
        return frozenlist([])
    return freeze(obj)
